from AccessControl import ModuleSecurityInfo
from Acquisition import aq_base
from Interface import Implements
from OFS.CopySupport import CopySource

from Products.CMFCore.utils import getToolByName
from Products.CMFCore.PortalFolder import PortalFolder
from Products.Archetypes.ArchetypeTool import listTypes

modulesec = ModuleSecurityInfo('Products.Relations.utils')
modulesec.declarePublic('isReferenceable')

class AllowedTypesByIface:
    """An approach to restrict allowed content types in a container by
    the interfaces they implement.

    Notice that extending this class means surpassing allowed_content_types,
    filter_content_types etc in the FTI, while we are still concerned about
    security."""

    # XXX: This class depends heavily on implementation details in CMF's
    #      PortalFolder. We need to eventually implement our own FTI, which
    #      should give us much cleaner code.

    allowed_interfaces = () # Don't allow anything, subclasses overwrite!

    def allowedContentTypes(self):
        """Redefines CMF PortalFolder's allowedContentTypes."""
        return self._getTIsByInterfaces(self.allowed_interfaces)

    def invokeFactory(self, type_name, id, RESPONSE = None, *args, **kwargs):
        """Invokes the portal_types tool.

        Overrides PortalFolder.invokeFactory."""
        goodType = self._portalTypeImplementsOneOfInterfaces
        if not goodType(type_name, self.allowed_interfaces):
            raise ValueError, "Type %s does not implement any of %s." % \
                  (type_name, self.allowed_interfaces)
        
        pt = getToolByName(self, 'portal_types')
        args = (type_name, self, id, RESPONSE) + args
        # XXX: PloneFolder returns an id. Do we really need to do that, too?
        new_id = pt.constructContent(*args, **kwargs)
        if not new_id: new_id = id
        return new_id

    def _verifyObjectPaste(self, object, validate_src=1):
        """Overrides PortalFolder._verifyObjectPaste."""
        # What we do here is trick PortalFolder._verifyObjectPaste in its check
        # for allowed content types. We make our typeinfo temporarily
        # unavailable.
        pt = getToolByName(self, 'portal_types')
        tmp_name = '%s_TMP' % self.portal_type
        ti = pt.getTypeInfo(self.portal_type)
        pt.manage_delObjects([self.portal_type])
        value = PortalFolder._verifyObjectPaste(self, object, validate_src)
        pt._setObject(self.portal_type, ti)
        return value

    def _getTIsByInterfaces(self, ifaces):
        """Returns a list of type info objects of available (AT) portal types
        implementing iface."""
        pt = getToolByName(self, 'portal_types')
        value = []
        for data in listTypes():
            klass = data['klass']
            for iface in self.allowed_interfaces:
                if iface.isImplementedByInstancesOf(klass):
                    ti = pt.getTypeInfo(data['portal_type'])
                    if ti is not None and ti.isConstructionAllowed(self):
                        value.append(ti)
        return value

    def _portalTypeImplementsOneOfInterfaces(self, portal_type, ifaces):
        """Does the given portal_type implement one of the given interfaces?"""
        klass = self._getClassByPortalType(portal_type)
        for iface in ifaces:
            if iface.isImplementedByInstancesOf(klass):
                return 1

    def _getClassByPortalType(self, name):
        for data in listTypes():
            if data['portal_type'] == name:
                return data['klass']
        raise ValueError, "No such portal type: %s" % name
        

def getPortalTypesByInterfaces(context, allowedIfaces):
    """Returns a list of portal type names of which the types implement any
    of the given interfaces.

    Note that allowedIfaces are strings, not Interfaces."""
    value = []
    tool = getToolByName(context, 'archetype_tool')
    for data in tool.listRegisteredTypes():
        kifaces = data['klass'].__implements__
        # We now flatten the interfaces and get their unqualified names,
        # e.g. <Products.FooBar.interfaces.ISpam> becomes "ISpam"
        kifaces = [str(iface.__name__)
                   for iface in Implements.flattenInterfaces((kifaces,))]
        if [iface for iface in allowedIfaces if iface in kifaces]:
            value.append(data['portal_type'])

    return value

def getReferenceableTypes(context):
    """Return a list of portal type strings, with all types in the
    types tool that are referenceable."""
    attool = getToolByName(context, 'archetype_tool')
    pttool = getToolByName(context, 'portal_types')
    attypes = [d['portal_type'] for d in attool.listRegisteredTypes()]
    return [t for t in pttool.listContentTypes() if t in attypes]
    

def isReferenceable(context):
    if getattr(aq_base(context), 'isReferenceable', None) is not None:
        return True
    else:
        return False

    
